/*
Copyright (C) 2013-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// utilities.c -- Baker: utilities with a layer of separation


#include "utilities.h" // This will include quakedef.h for us if applicable

#ifdef QUAKE_GAME
#include "quakedef.h"
#include "utilities.h"

#endif // QUAKE_GAME

void Utilities_Init (void)
{
//	Pak_Init (qfunction_set);

#if 0
	Wad2_Init (qfunction_set);
	Downloads_Init (qfunction_set);
	Zip_Init (qfunction_set);
#endif

	Cmd_AddCommands (Utilities_Init);
}

const char *cmds[] = {"add", "compress", "extract", "help", "list", "rem", "replace", "unzip", "zip", NULL};
enum arg_e {arg_add, arg_compress, arg_extract, arg_help, arg_list, arg_rem, arg_replace, arg_unzip, arg_zip, max_args};

const char *arg_msg[] =
{
	"<file.pak> <file in pak> <source url>", // add
	"<file.pak>", // compress
	"<file.pak> <file in pak> <extract destination>", // extract
	"", // help
	"<file.pak>", // list
	"<file.pak> <file in pak>", // remove
	"<file.pak> <file in pak> <replace with file path>", // replace
	"<file.pak> <destination dir>", //unzip
	"<file.pak to create> <folder to pak>", // zip
	NULL,
};


int args_required[] =
{
	3, // add pakurl pak_file_name filenameurl
	1, // compress pakurl
	3, // extract pakurl filename
	0, // help
	1, // list pakurl
	2, // remove pakurl filename
	3, // replace pakurl filename filenameurl
	2, // unzip pakurl workingdir
	2, // zip pakurl compressdir
	0,
};


// compress, extract, rem, replace will remain unlisted as to not confuse users.
void Pak_Command_f (void)
{
	// replace file, delete file, compress


	if (Cmd_Argc() >= 2)
	{
		const char *parm2 = Cmd_Argv(1);
		const char *_pakurl = Cmd_Argc() >= 3 ? Cmd_Argv (2) : NULL;
		const char *arg1 = Cmd_Argc() >= 4 ? Cmd_Argv (3) : NULL;
		const char *arg2 = Cmd_Argc() >=5 ? Cmd_Argv(4) : NULL;

		char pakurl[MAX_OSPATH];
		char fileurl[MAX_OSPATH];
		char destfolder[MAX_OSPATH];
		char srcfolder[MAX_OSPATH];
		char inside_pak_filename[MAX_OSPATH];

		int cmd_num = String_To_Array_Index (parm2, cmds);
		int numargs = Cmd_Argc() - 2;
		int i;
		cbool validnum = cmd_num < max_args;
		cbool meets_args = validnum ? args_required[cmd_num] == numargs : false;

		if (!validnum)
		{
			Con_Printf ("Invalid pak command %s\n", parm2);
			return;
		}

		if (!meets_args)
		{
			Con_Printf ("Wrong number of arguments for \"%s\": needs %i, have %i\n", parm2, args_required[cmd_num], numargs);
			Con_Printf ("Usage: %s %s %s\n", Cmd_Argv(0), parm2, arg_msg[cmd_num]);
			return;
		}

		c_strlcpy (pakurl, _pakurl);
		File_URL_Edit_SlashesForward_Like_Unix (pakurl);

		if (cmd_num != arg_zip && !File_Exists (pakurl))
		{
			Con_Printf ("File \"%s\" doesn't exist\n", _pakurl); // Print the unmodified version
			return;
		}

		switch (cmd_num)
		{
		case arg_add:  // sorts and rewrites (where nulled entry space is recovered) (can just sort the header)
	
			c_strlcpy (inside_pak_filename, arg1); File_URL_Edit_SlashesForward_Like_Unix (inside_pak_filename);
			c_strlcpy (fileurl, arg2); File_URL_Edit_SlashesForward_Like_Unix (fileurl);

			// Make sure the new file exists
			if (!File_Exists (fileurl))
			{
				Con_Printf ("New file \"%s\" does not exist\n", arg2);
				return;
			}

			if (Pak_Has_File(pakurl, inside_pak_filename))
			{
				Con_Printf ("A file named \"%s\" is already in the pak\n", inside_pak_filename);
				return;
			}

			// We already know the pak exists so lets get to it
			Pak_Add_File (pakurl, inside_pak_filename, fileurl /*source*/);

			Recent_File_Set_FullPath (pakurl);
			return;


		case arg_compress:  // sorts and rewrites (where nulled entry space is recovered) (can just sort the header)

			if (Pak_Compress (pakurl))
				Con_Printf ("Pak compress successful\n");

			else Con_Printf ("Pak compress failed\n");

			Recent_File_Set_FullPath (pakurl);
			return;

		case arg_extract:	// extract file to <dest> (final arg is not folder but entire path of file to write!!!!)

			//File_URL_Edit_SlashesForward_Like_Unix (pakurl);
			c_strlcpy (inside_pak_filename, arg1); File_URL_Edit_SlashesForward_Like_Unix (inside_pak_filename);
			c_strlcpy (fileurl, arg2); File_URL_Edit_SlashesForward_Like_Unix (fileurl);

			// Validate destination folder, making sure it exists
			c_strlcpy (destfolder, fileurl);
			File_URL_Edit_Reduce_To_Parent_Path (destfolder);

			if (!File_Exists (destfolder))
			{
				Con_Printf ("Destination folder for \"%s\" does not exist\n", arg2);
				return;
			}

			if (!File_Is_Folder(destfolder))
			{
				Con_Printf ("Destination folder for \"%s\" is a file\n", arg2);
				return;
			}

			// Validate file to extract exists
			c_strlcpy (inside_pak_filename, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (inside_pak_filename);

			if (!Pak_Has_File (pakurl, inside_pak_filename))
			{
				Con_Printf ("File \"%s\" does not exist in pak \"%s\"\n", inside_pak_filename, pakurl);
				return;
			}

			Pak_Extract_File (pakurl, inside_pak_filename, fileurl /*dest*/);

			Recent_File_Set_FullPath (fileurl);
			return;

		case arg_rem:		// remove a file <delete a file inside pak by zeroing out entry>

			// Validate file to zero out inside pak exists
			c_strlcpy (inside_pak_filename, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (inside_pak_filename);

			if (!Pak_Has_File (pakurl, inside_pak_filename))
			{
				Con_Printf ("File \"%s\" does not exist in pak \"%s\"\n", inside_pak_filename, pakurl);
				return;
			}

			Pak_Remove_File (pakurl, inside_pak_filename);
			return;

		case arg_replace:	return; // replace a file in pak <if larger, null out entry and write another one>

			// Validate source
			c_strlcpy (fileurl, arg2);
			File_URL_Edit_SlashesForward_Like_Unix (fileurl);

			if (!File_Exists (fileurl))
			{
				Con_Printf ("File source \"%s\" does not exist\n", fileurl);
				return;
			}

			// Validate requested file to replace exists
			c_strlcpy (inside_pak_filename, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (inside_pak_filename);

			if (!Pak_Has_File (pakurl, inside_pak_filename))
			{
				Con_Printf ("File \"%s\" does not exist in pak \"%s\"\n", inside_pak_filename, pakurl);
				return;
			}
#pragma message ("Baker: Has pak replace ever been tested?")
			Pak_Replace_File (pakurl, inside_pak_filename /*replaced*/ , fileurl);
			return;

		case arg_unzip: // arg1 is destination

			// Validate destination
			c_strlcpy (destfolder, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (destfolder);

			if (!File_Exists (destfolder))
			{
				char parentpath[MAX_OSPATH];
				c_strlcpy (parentpath, destfolder);
				File_URL_Edit_Reduce_To_Parent_Path (parentpath);

				if (!File_Exists(parentpath))
				{
					Con_Printf ("Destination \"%s\" does not exist\n", destfolder);
					return;
				}
				// Ok to make just a subfolder
			}

			i = Pak_Unzip (pakurl, destfolder);

			Con_Printf ("%i files extracted\n", i);
			Recent_File_Set_FullPath (destfolder);
			return;

		case arg_zip: // arg1 is source

			// Validate source
			c_strlcpy (srcfolder, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (srcfolder);

			if (!File_Exists (srcfolder))
			{
				Con_Printf ("Source folder \"%s\" does not exist\n", srcfolder);
				return;
			}

			Pak_Zip_Folder (pakurl, srcfolder);

			Recent_File_Set_FullPath (pakurl);
			return;

		case arg_help: // list current mappings

			Con_Printf ("\n");
			Con_Printf ("Usage: %s <0-99> <yourfile.mp3>\nwhere yourfile.mp3 is in [gamedir]/music folder\n", Cmd_Argv (0));
			Con_Printf ("\n");
			Con_Printf ("Usage: %s list  - print list of contents\n", Cmd_Argv (0));
			Con_Printf ("Usage: %s zip <folder_to_zip> <output.pak> - resets everything\n", Cmd_Argv (0));
			Con_Printf ("Usage: %s unzip <input.pak> <folder_unzip_location>\n", Cmd_Argv (0));
			Con_Printf ("\n");
			Con_Printf ("Examples:\n");
			Con_Printf ("\n");
			Con_Printf ("%s unzip c:\\quake\\id1\\pak.0.pak c:\\quake2\n", Cmd_Argv (0));
			Con_Printf ("%s zip c:\\quake\\id1\\maps c:\\quake\\id1\\mymaps.pak\n", Cmd_Argv (0));
			Con_Printf ("\n");
			Con_Printf ("Filenames should avoid spaces and only use alphanumeric\n");
			Con_Printf ("characters and the underscore '_'.\n");
			Con_Printf ("\n");
			Con_Printf ("Type 'folder' to access current gamedir.\n");
			Con_Printf ("\n");

			return;

		case arg_list: // list current mappings

			Con_Printf ("\n");
			Con_Printf ("Files in pak \"%s\":\n", _pakurl);
			Con_Printf ("\n");

			Pak_List_Print (pakurl);

			Con_Printf ("\n");

			Con_Printf ("Savings by compressing pak is %u\n", (unsigned int) Pak_Is_Compressable(pakurl));

			Recent_File_Set_FullPath (pakurl);
			return;

		} // End of switch

	} // End of args >=2

	// with no parameters or invalid parameters ends up displaying help
	Con_Printf ("\n");
	Con_Printf ("Usage: %s {help|list|unzip|zip}\n", Cmd_Argv (0));
	Con_Printf ("Type '%s help' for examples and detail\n", Cmd_Argv (0));
	Con_Printf ("\n");

}

// compress, extract, rem, replace will remain unlisted as to not confuse users.
// compress, extract, rem, replace will remain unlisted as to not confuse users.
void Zip_Command_f (void)
{
	// replace file, delete file, compress


	if (Cmd_Argc() >= 2)
	{
		const char *parm2 = Cmd_Argv(1);
		const char *_zipurl = Cmd_Argc() >= 3 ? Cmd_Argv (2) : NULL;
		const char *arg1 = Cmd_Argc() >= 4 ? Cmd_Argv (3) : NULL;
		const char *arg2 = Cmd_Argc() >=5 ? Cmd_Argv(4) : NULL;

		char zipurl[MAX_OSPATH];
		char fileurl[MAX_OSPATH];
		char destfolder[MAX_OSPATH];
		char srcfolder[MAX_OSPATH];
		char inside_zip_filename[MAX_OSPATH];

		int cmd_num = String_To_Array_Index (parm2, cmds);
		int numargs = Cmd_Argc() - 2;
		int i;
		cbool validnum = cmd_num < max_args;
		cbool meets_args = validnum ? args_required[cmd_num] == numargs : false;

		if (!validnum)
		{
			Con_Printf ("Invalid zip command %s\n", parm2);
			return;
		}

		if (!meets_args)
		{
			Con_Printf ("Wrong number of arguments for \"%s\": needs %i, have %i\n", parm2, args_required[cmd_num], numargs);
			Con_Printf ("Usage: %s %s %s\n", Cmd_Argv(0), parm2, arg_msg[cmd_num]);
			return;
		}

		c_strlcpy (zipurl, _zipurl);
		File_URL_Edit_SlashesForward_Like_Unix (zipurl);

		if (cmd_num != arg_zip && !File_Exists (zipurl))
		{
			Con_Printf ("File \"%s\" doesn't exist\n", _zipurl); // Print the unmodified version
			return;
		}

		switch (cmd_num)
		{
		case arg_add:  // sorts and rewrites (where nulled entry space is recovered) (can just sort the header)
			Con_Printf ("Unsupported at this time\n");
			return;


		case arg_compress:  // sorts and rewrites (where nulled entry space is recovered) (can just sort the header)

			Con_Printf ("Unsupported at this time\n");
			return;

		case arg_extract:	// extract file to <dest> (final arg is not folder but entire path of file to write!!!!)

			//File_URL_Edit_SlashesForward_Like_Unix (zipurl);
			c_strlcpy (inside_zip_filename, arg1); File_URL_Edit_SlashesForward_Like_Unix (inside_zip_filename);
			c_strlcpy (fileurl, arg2); File_URL_Edit_SlashesForward_Like_Unix (fileurl);

			// Validate destination folder, making sure it exists
			c_strlcpy (destfolder, fileurl);
			File_URL_Edit_Reduce_To_Parent_Path (destfolder);

			if (!File_Exists (destfolder))
			{
				Con_Printf ("Destination folder for \"%s\" does not exist\n", arg2);
				return;
			}

			if (!File_Is_Folder(destfolder))
			{
				Con_Printf ("Destination folder for \"%s\" is a file\n", arg2);
				return;
			}

			// Validate file to extract exists
			c_strlcpy (inside_zip_filename, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (inside_zip_filename);

			if (!Zip_Has_File (zipurl, inside_zip_filename))
			{
				Con_Printf ("File \"%s\" does not exist in zip \"%s\"\n", inside_zip_filename, zipurl);
				return;
			}

			Zip_Extract_File (zipurl, inside_zip_filename, fileurl /*dest*/);

			Recent_File_Set_FullPath (fileurl);
			return;

		case arg_rem:		// remove a file <delete a file inside zip by zeroing out entry>

			Con_Printf ("Unsupported at this time\n");
			return;

		case arg_replace:	return; // replace a file in zip <if larger, null out entry and write another one>

			Con_Printf ("Unsupported at this time\n");
			return;

		case arg_unzip: // arg1 is destination

			// Validate destination
			c_strlcpy (destfolder, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (destfolder);

			if (!File_Exists (destfolder))
			{
				char parentpath[MAX_OSPATH];
				c_strlcpy (parentpath, destfolder);
				File_URL_Edit_Reduce_To_Parent_Path (parentpath);

				if (!File_Exists(parentpath))
				{
					Con_Printf ("Destination \"%s\" does not exist\n", destfolder);
					return;
				}
				// Ok to make just a subfolder
			}

			i = Zip_Unzip (zipurl, destfolder);

			Con_Printf ("%i files extracted\n", i);
			Recent_File_Set_FullPath (destfolder);
			return;

		case arg_zip: // arg1 is source

			// Validate source
			c_strlcpy (srcfolder, arg1);
			File_URL_Edit_SlashesForward_Like_Unix (srcfolder);

			if (!File_Exists (srcfolder))
			{
				Con_Printf ("Source folder \"%s\" does not exist\n", srcfolder);
				return;
			}

			Zip_Zip_Folder (zipurl, srcfolder);

			Recent_File_Set_FullPath (zipurl);
			return;

		case arg_help: // list current mappings

			Con_Printf ("\n");
			Con_Printf ("Usage: %s <0-99> <yourfile.mp3>\nwhere yourfile.mp3 is in [gamedir]/music folder\n", Cmd_Argv (0));
			Con_Printf ("\n");
			Con_Printf ("Usage: %s list  - print list of contents\n", Cmd_Argv (0));
			Con_Printf ("Usage: %s zip <folder_to_zip> <output.zip> - resets everything\n", Cmd_Argv (0));
			Con_Printf ("Usage: %s unzip <input.zip> <folder_unzip_location>\n", Cmd_Argv (0));
			Con_Printf ("\n");
			Con_Printf ("Examples:\n");
			Con_Printf ("\n");
			Con_Printf ("%s unzip c:\\quake\\id1\\zip.0.zip c:\\quake2\n", Cmd_Argv (0));
			Con_Printf ("%s zip c:\\quake\\id1\\maps c:\\quake\\id1\\mymaps.zip\n", Cmd_Argv (0));
			Con_Printf ("\n");
			Con_Printf ("Filenames should avoid spaces and only use alphanumeric\n");
			Con_Printf ("characters and the underscore '_'.\n");
			Con_Printf ("\n");
			Con_Printf ("Type 'folder' to access current gamedir.\n");
			Con_Printf ("\n");

			return;

		case arg_list: // list current mappings

			Con_Printf ("\n");
			Con_Printf ("Files in zip \"%s\":\n", _zipurl);
			Con_Printf ("\n");

			Zip_List_Print (zipurl);

			Con_Printf ("\n");

			Recent_File_Set_FullPath (zipurl);
			return;

		} // End of switch

	} // End of args >=2

	// with no parameters or invalid parameters ends up displaying help
	Con_Printf ("\n");
	Con_Printf ("Usage: %s {help|list|unzip|zip}\n", Cmd_Argv (0));
	Con_Printf ("Type '%s help' for examples and detail\n", Cmd_Argv (0));
	Con_Printf ("\n");

}


void Dir_Command_f (void)
{
	clist_t *dir_contents = File_List_Recursive_Alloc (com_gamedir);
	clist_t *cursor;
	double filesizes;
	size_t cursize;
	const char *comparestring = NULL;

	if (Cmd_Argc () == 2)
		comparestring = Cmd_Argv (1);

	for (cursor = dir_contents, filesizes = 0; cursor; cursor = cursor->next)
	{
		// Skip disqualified matches
		if (comparestring && !wildmultcmp (comparestring, File_URL_SkipPath (cursor->name)) )
			continue;
		cursize = File_Length(cursor->name),
		Con_Printf ("%s %u\n", cursor->name, (unsigned int) cursize); // C89
		filesizes += cursize;
	}

	Con_Printf ("Sizes combined %f\n", filesizes);
	List_Free (&dir_contents);
}


void Download_Command_f (void)
{
	if (Cmd_Argc() == 2)
	{ //http://quake-1.com/mods/barnak.zip
		const char *urlname = Cmd_Argv(1);
		const char *fullurl = basedir_to_url(File_URL_SkipPath(urlname));
		
		Download_To_File(urlname, fullurl, NULL);
		Recent_File_Set_FullPath (fullurl);
		
	} else Con_Printf ("Need 1 argument, the web site address with http:// in it");
}


void OpenD_Command_f (void)
{
//	Folder_Open
	Con_Printf ("%s\n", System_Dialog_Open_Type ("Open a file", "c:/quake/cda", ".pak,.png"));
}

void SaveD_Command_f (void)
{
}






